from pathlib import Path


class FileRangeTransBorderError(Exception):
    """文件范围超出文件大小错误"""

    def __init__(
            self,
            file_path: str | Path,
            correct_size: int,
            message: str = 'The file size is out of bounds',
    ):
        self.message = message
        self.file_path = Path(file_path)
        self.file_size = self.file_path.stat().st_size
        self.correct_size = correct_size
        error_msg = f'{self.file_path.name} Error: {message} | server size {correct_size} not equal local size {self.file_size}'
        super().__init__(error_msg)


class NotSliceSizeError(Exception):
    """没有文件大小的切片下载错误"""

    def __init__(
            self,
            file_path: str | Path,
            message: str = 'The file size is not divisible by the slice size'):
        self.message = message
        self.file_path = Path(file_path)
        error_msg = f'{self.file_path.name} Error: {message}'
        super().__init__(error_msg)



from yundownload.core import Auth, Proxies, Retry, Result
from yundownload.request import Request
from yundownload.pool import AsyncDownloadPools, DownloadPools
from yundownload.utils import (get_stream_server_size,
                               get_head_server_size,
                               async_get_stream_server_size,
                               async_get_head_server_size,
                               merge_file,
                               async_merge_file,
                               async_event_loop_select,
                               convert_bytes_per_second)
from yundownload.logger import show_log, write_log, logger

from yundownload.cli import cli, get_version, render_ui

__version__ = get_version()



from dataclasses import dataclass
from enum import IntFlag
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from yundownload.request import Request


@dataclass
class Auth:
    username: str
    password: str


@dataclass
class Proxies:
    http: str
    https: str


@dataclass
class Retry:
    retry: int = 1
    retry_delay: int = 10
    retry_connect: int = 5


class Status(IntFlag):
    """
    下载状态
    """
    WAIT = 1
    FAIL = 2
    SUCCESS = 4
    SLICE = 8
    EXIST = 16


@dataclass
class Result:
    """
    下载结果
    """
    status: Status
    request: 'Request'
    error: Exception = None



import logging
from logging.handlers import RotatingFileHandler


class CustomFormatter(logging.Formatter):
    def format(self, record):
        log_format = f'[%(asctime)s] [%(levelname)s] >> %(message)s'
        formatter = logging.Formatter(log_format)
        return formatter.format(record)


class CustomHandler(logging.StreamHandler):
    def emit(self, record):
        super().emit(record)


def show_log():
    handler = CustomHandler()
    handler.setFormatter(CustomFormatter())
    logger.addHandler(handler)


def write_log(log_path: str):
    handler = RotatingFileHandler(log_path, maxBytes=100 * 1024 * 1024, encoding='utf-8')
    handler.setFormatter(CustomFormatter())
    logger.addHandler(handler)


logger = logging.getLogger('yundownload')
logger.setLevel(logging.INFO)



import asyncio
import signal
import time
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor, wait, Future
from typing import TYPE_CHECKING, Callable, Optional, Awaitable

import httpx
from httpx import Client, AsyncClient

from yundownload.core import Retry, Result, Status, Proxies, Auth
from yundownload.downloader import stream_downloader, slice_downloader, async_stream_downloader, async_slice_downloader
from yundownload.logger import logger

if TYPE_CHECKING:
    from yundownload.request import Request

"""
uvloop 优化协程
"""


class BaseDP:
    _retry: Retry = Retry()
    _future_list = []

    @abstractmethod
    def push(self, item: 'Request'):
        pass

    @abstractmethod
    def _task_fail(self, result: Result):
        pass

    @abstractmethod
    def _task_handle(self, item: 'Request'):
        pass

    @abstractmethod
    def _task_start(self, item: 'Request', func: Callable[[Client | AsyncClient, 'Request'], Result]):
        pass

    @abstractmethod
    def close(self):
        pass


class DownloadPools(BaseDP):

    def __init__(
            self,
            max_workers: int = 5,
            timeout: int = 20,
            retry: Retry = Retry(),
            verify: bool = True,
            proxies: Optional['Proxies'] = None,
            auth: Optional['Auth'] = None,
    ) -> None:
        self._retry = retry
        self._thread_pool: ThreadPoolExecutor = ThreadPoolExecutor(max_workers=max_workers)
        self.client = httpx.Client(
            transport=httpx.HTTPTransport(
                retries=self._retry.retry_connect
            ),
            timeout=timeout,
            follow_redirects=True,
            verify=verify,
            mounts=proxies if proxies is None else {
                "http://": httpx.HTTPTransport(proxy=proxies.http),
                "https://": httpx.HTTPTransport(proxy=proxies.https),
            },
            auth=None if auth is None else httpx.BasicAuth(
                username=auth.username,
                password=auth.password
            )
        )

    def _task_start(self, item: 'Request',
                    func: Callable[['Client', 'Request', Callable], Result]) -> Result:
        return func(self.client, item, self._pool_submit)

    def push(self, item: 'Request'):
        future: Future = self._pool_submit(self._task_handle, item)
        self._future_list.append(future)
        pass

    @property
    def _pool_submit(self):
        return self._thread_pool.submit

    def _task_handle(self, item: 'Request') -> Result:
        err = None
        slice_flag = False
        result = None
        for i in range(self._retry.retry):
            try:
                if not slice_flag:
                    item.stat.close()
                    result = self._task_start(item, stream_downloader)
                    if result.status == Status.SLICE:
                        slice_flag = True
                if slice_flag:
                    item.stat.close()
                    result = self._task_start(item, slice_downloader)
                item.status = result.status
                return result
            except Exception as e:
                err = e
                logger.warning(
                    f"retrying... {i} of {self._retry.retry}, error msg: {e} line: {e.__traceback__.tb_lineno}")
                time.sleep(self._retry.retry_delay)
                continue
        result = Result(
            status=Status.FAIL,
            request=item,
            error=err
        )
        self._task_fail(result)
        item.status = result.status
        return result

    def _task_fail(self, result: Result):
        result.request.error_callback(result)

    def results(self):
        return [future.result() for future in self._future_list]

    def close(self):
        wait(self._future_list)
        self._thread_pool.shutdown(wait=True)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class AsyncDownloadPools(BaseDP):

    def __init__(
            self,
            max_workers: int = 5,
            timeout: int = 20,
            retry: Retry = Retry(),
            verify: bool = True,
            proxies: Optional['Proxies'] = None,
            auth: Optional['Auth'] = None,
    ) -> None:
        self._retry = retry
        self._semaphore: asyncio.Semaphore = asyncio.Semaphore(max_workers)
        self.client = AsyncClient(
            transport=httpx.AsyncHTTPTransport(
                retries=self._retry.retry_connect
            ),
            timeout=timeout,
            follow_redirects=True,
            verify=verify,
            mounts=proxies if proxies is None else {
                "http://": httpx.AsyncHTTPTransport(proxy=proxies.http),
                "https://": httpx.AsyncHTTPTransport(proxy=proxies.https),
            },
            auth=None if auth is None else httpx.BasicAuth(
                username=auth.username,
                password=auth.password
            )
        )

    async def _task_start(self, item: 'Request',
                          func: Callable[[AsyncClient, 'Request', asyncio.Semaphore], Awaitable[Result]]):
        return await func(self.client, item, self._semaphore)

    async def push(self, item: 'Request'):
        future = asyncio.create_task(self._task_handle(item))
        self._future_list.append(future)

    async def _task_handle(self, item: 'Request') -> Result:
        err = None
        slice_flag = False
        result = None
        for i in range(self._retry.retry):
            try:
                if not slice_flag:
                    item.stat.close()
                    result = await self._task_start(item, async_stream_downloader)
                    if result.status == Status.SLICE:
                        slice_flag = True
                if slice_flag:
                    item.stat.close()
                    result = await self._task_start(item, async_slice_downloader)
                item.status = result.status
                return result
            except Exception as e:
                raise e
                err = e
                logger.warning(
                    f"retrying... {i} of {self._retry.retry}, error msg: {e} line: {e.__traceback__.tb_lineno}")
                time.sleep(self._retry.retry_delay)
                continue
        result = Result(
            status=Status.FAIL,
            request=item,
            error=err
        )
        logger.error(err, exc_info=True)
        await self._task_fail(result)
        item.status = result.status
        return result

    async def _task_fail(self, result: Result):
        await result.request.error_callback(result)

    async def results(self):
        return await asyncio.gather(*self._future_list)

    async def close(self):
        await self.results()
        await self.client.aclose()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



import asyncio
import time
import platform
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import aiofiles
from httpx import Client, AsyncClient

from yundownload.core import Status
from yundownload.exception import FileRangeTransBorderError
from yundownload.logger import logger

if TYPE_CHECKING:
    from yundownload.request import Request


def check_range_transborder(
        client: Client,
        request: 'Request'
) -> None:
    try:
        correct_size = get_head_server_size(client, request)
    except:
        correct_size = get_stream_server_size(client, request)
    if correct_size < request.save_path.stat().st_size:
        if request.transborder_delete:
            request.save_path.unlink()
        raise FileRangeTransBorderError(
            file_path=request.save_path,
            correct_size=correct_size,
        )


async def async_check_range_transborder(
        client: AsyncClient,
        request: 'Request'
) -> None:
    try:
        correct_size = await async_get_head_server_size(client, request)
    except:
        correct_size = await async_get_stream_server_size(client, request)
    if correct_size < request.save_path.stat().st_size:
        if request.transborder_delete:
            request.save_path.unlink()
        raise FileRangeTransBorderError(
            file_path=request.save_path,
            correct_size=correct_size,
        )


def get_stream_server_size(
        client: Client,
        request: 'Request'
) -> int:
    with client.stream(
            method=request.method,
            url=request.url,
            params=request.params,
            data=request.data,
            headers=request.headers,
            cookies=request.cookies,
            auth=request.auth,
            timeout=request.timeout,
            follow_redirects=request.follow_redirects,
    ) as response:
        response.raise_for_status()
        return int(response.headers.get('Content-Length', 0))


async def async_get_stream_server_size(
        client: AsyncClient,
        request: 'Request'
) -> int:
    async with client.stream(
            method=request.method,
            url=request.url,
            params=request.params,
            data=request.data,
            headers=request.headers,
            cookies=request.cookies,
            auth=request.auth,
            timeout=request.timeout,
            follow_redirects=request.follow_redirects,
    ) as response:
        response.raise_for_status()
        return int(response.headers.get('Content-Length', 0))


def get_head_server_size(
        client: Client,
        request: 'Request'
) -> int:
    response = client.head(
        url=request.url,
        params=request.params,
        headers=request.headers,
        cookies=request.cookies,
        auth=request.auth,
        timeout=request.timeout,
        follow_redirects=request.follow_redirects,
    )
    response.raise_for_status()
    return int(response.headers.get('Content-Length'))


async def async_get_head_server_size(
        client: AsyncClient,
        request: 'Request'
) -> int:
    response = await client.head(
        url=request.url,
        params=request.params,
        headers=request.headers,
        cookies=request.cookies,
        auth=request.auth,
        timeout=request.timeout,
        follow_redirects=request.follow_redirects,
    )
    response.raise_for_status()
    return int(response.headers.get('Content-Length'))


def retry(retry_num: int, retry_delay: int):
    def wrapper(func):
        def inner(*args, **kwargs):
            err = None
            for i in range(retry_num):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    err = e
                    logger.warning(f"An error has occurred, retrying, {i} of {retry_num}, error information: {e}")
                    time.sleep(retry_delay)
                    continue
            raise err

        return inner

    return wrapper


def merge_file(slice_flag: str, save_path: Path):
    """
    合并文件
    :param slice_flag: 切片文件标识
    :param save_path: 保存路径
    :return:
    """
    logger.info(f"Start merging files, file path: {save_path}")
    slice_files = list(save_path.parent.glob(f'*{slice_flag}*.ydstf'))
    slice_files.sort(key=lambda x: int(x.stem.split('--')[1]))
    with save_path.open('wb') as wf:
        for file in slice_files:
            logger.info(f'merge chunk: [{file}]')
            with file.open('rb') as rf:
                while True:
                    chunk = rf.read(4096)
                    if not chunk:
                        break
                    wf.write(chunk)

    for file in slice_files:
        file.unlink()
    logger.info(f"Merge file success, file path: {save_path}")


async def async_merge_file(slice_flag: str, save_path: Path):
    """
    合并文件
    :param slice_flag: 切片文件标识
    :param save_path: 保存路径
    :return:
    """
    logger.info(f"Start merging files, file path: {save_path}")
    slice_files = list(save_path.parent.glob(f'*{slice_flag}*.ydstf'))
    slice_files.sort(key=lambda x: int(x.stem.split('--')[1]))
    async with aiofiles.open(save_path, 'wb') as wf:
        for file in slice_files:
            logger.info(f'merge chunk: [{file}]')
            async with aiofiles.open(file, 'rb') as rf:
                while True:
                    chunk = await rf.read(4096)
                    if not chunk:
                        break
                    await wf.write(chunk)
    for file in slice_files:
        file.unlink()
    logger.info(f"Merge file success, file path: {save_path}")


async def async_event_loop_select():
    if platform.system() == 'Windows':
        warnings.warn('Windows system does not support uvloop, use default event loop')
        return
    else:
        import uvloop
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        loop = uvloop.new_event_loop()
        asyncio.set_event_loop(loop)


def convert_bytes_per_second(bytes_per_second):
    """将字节数/秒转换为自适应单位"""
    units = ['B/s', 'KB/s', 'MB/s', 'GB/s', 'TB/s']
    index = 0

    while bytes_per_second >= 1024 and index < len(units) - 1:
        bytes_per_second /= 1024.0
        index += 1

    return f"{bytes_per_second:.2f} {units[index]}"


def status_to_str(status: Status):
    return {
        Status.SUCCESS: 'Success',
        Status.FAIL: 'Fail',
        Status.EXIST: 'Exist',
        Status.WAIT: 'Wait',
        Status.SLICE: 'Slice'
    }[status]



import argparse
import time
from importlib.metadata import version
from typing import Iterator
from typing import List

from rich.console import Console
from rich.progress import Progress
from rich.table import Table

from yundownload.core import Retry, Status
from yundownload.pool import DownloadPools
from yundownload.request import Request
from yundownload.utils import convert_bytes_per_second


def render_ui(requests: List[Request]):
    table = Table(title="Yun Downloader", show_lines=True, expand=True)
    table.add_column("Filename", justify="right", style="cyan", no_wrap=True)
    table.add_column("URI", style="magenta")
    for request in requests:
        table.add_row(request.save_path.name, request.url)

    console = Console()

    with Progress(console=console) as progress:
        for request in requests:
            request.meta['task'] = progress.add_task(f"[red]{request.save_path.name}...", total=1)

        is_ok = False
        while not is_ok:
            console.clear()
            console.print(table)
            speed = 0
            for request in requests:
                percentage = request.stat.percentage
                progress.update(request.meta['task'], completed=percentage)
                speed += request.stat.speed
            console.print(f"[bold green]Speed: {convert_bytes_per_second(speed)}")
            is_ok = all(bool((Status.SUCCESS | Status.FAIL | Status.EXIST) & request.status) for request in requests)
            time.sleep(2.5)


def get_version():
    return version("yundownload")


def load_file(file: str) -> Iterator:
    with open(file, 'r', encoding='utf-8') as f:
        for line in iter(f.readline, ''):
            save_path, url = line.strip().split('<yfd>')
            yield Request(url=url, save_path=save_path)


def cli() -> None:
    parser = argparse.ArgumentParser(description='Yun Downloader')
    subparsers = parser.add_subparsers(dest='subcommand')
    load_parser = subparsers.add_parser('load', help='Load a request')
    load_parser.add_argument('file', type=str, help='yfd file load')
    load_parser.set_defaults(help=load_parser.print_help)

    download = subparsers.add_parser('download', help='Download a file')
    download.add_argument('url', type=str, help='Download url')
    download.add_argument('save_path', type=str, help='Save path, including file name')
    download.add_argument('-mw', '--maxworker', default=8, type=int, help='Maximum concurrency')
    download.add_argument('-t', '--timeout', type=int, default=100, help='Timeout period')
    download.add_argument('-r', '--retry', type=int, default=1, help='Retry times')
    download.add_argument('--stream', action='store_true', default=False, help='Forced streaming')
    download.add_argument('-V', '--version', action='version', version=f'%(prog)s {get_version()}',
                          help='Display version')
    download.set_defaults(help=download.print_help)
    parser.set_defaults(help=parser.print_help)
    args = parser.parse_args()
    if args.subcommand == 'download':
        request = Request(
            url=args.url,
            save_path=args.save_path,
            timeout=args.timeout
        )
        if args.stream:
            request.stream = True
        with DownloadPools(max_workers=args.maxworker, retry=Retry(retry=args.retry)) as pool:
            pool.push(request)
    elif args.subcommand == 'load':
        with DownloadPools() as pool:
            requests = []
            for request in load_file(args.file):
                pool.push(request)
                requests.append(request)
            render_ui(requests)



from pathlib import Path
from typing import Literal, TYPE_CHECKING, Optional, Callable
from urllib.parse import urljoin

import httpx

from yundownload.core import Result, Status
from yundownload.logger import logger
from yundownload.stat import Stat

if TYPE_CHECKING:
    from yundownload.core import Auth


class Request:
    def __init__(
            self,
            url: str,
            save_path: str | Path,
            method: Literal["GET", "POST", "HEAD", "PUT", "DELETE"] = 'GET',
            params: dict = None,
            headers: dict = None,
            cookies: dict = None,
            data: dict = None,
            slice_threshold=500 * 1024 * 1024,  # 500M
            slice_size: int = 100 * 1024 * 1024,  # 100M
            auth: Optional['Auth'] = None,
            timeout: int = 20,
            follow_redirects: bool = True,
            stream_size: int = 1024,
            success_callback: Optional[Callable[['Result'], None]] = None,
            error_callback: Optional[Callable[['Result'], None]] = None
    ):
        self.url = url
        self.save_path = Path(save_path)
        self.stream_size = stream_size
        self.slice_size = slice_size
        self.correct_size = None
        self.method = method
        self.slice_threshold = slice_threshold
        self.params = params or {}
        self.meta = {}
        self.headers = {
            'User-Agent': 'Wget/1.12 (linux-gnu)',
            'Content-Encoding': 'identity',
            'Accept-Encoding': 'identity'
        }
        if headers is not None:
            self.headers.update(headers)
        self.cookies = cookies or {}
        self.data = data or {}
        self.transborder_delete = False
        self.auth = None if auth is None else httpx.BasicAuth(
            username=auth.username,
            password=auth.password
        )
        # 为 True 时，使用流式下载不采用分片
        self.stream = False
        self.timeout = timeout
        self.follow_redirects = follow_redirects
        if callable(success_callback) or success_callback is None:
            self._success_callback = success_callback
        else:
            logger.warning("callback is not callable")
            self._success_callback = None
        if callable(error_callback) or error_callback is None:
            self._error_callback = error_callback
        else:
            logger.warning("callback is not callable")
            self._error_callback = None
        self.stat = Stat(self)
        self.status: Optional[Status] = Status.WAIT

    def success_callback(self, result: 'Result'):
        self.stat.close()
        if self._success_callback is None:
            return
        return self._success_callback(result)

    async def asuccess_callback(self, result: 'Result'):
        self.stat.close()
        if self._success_callback is None:
            return
        return await self._success_callback(result)

    def error_callback(self, result: 'Result'):
        self.stat.close()
        if self._error_callback is None:
            return
        return self._error_callback(result)

    async def aerror_callback(self, result: 'Result'):
        self.stat.close()
        if self._error_callback is None:
            return
        return await self._error_callback(result)

    def join(self, other: str) -> str:
        return urljoin(self.url, other)



import asyncio
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Callable

import aiofiles
from httpx import Client, Response, AsyncClient

from yundownload.core import Result, Status
from yundownload.exception import NotSliceSizeError
from yundownload.logger import logger
from yundownload.utils import check_range_transborder, merge_file, async_check_range_transborder, async_merge_file

if TYPE_CHECKING:
    from yundownload.request import Request


def stream_downloader(
        client: Client,
        request: 'Request',
        poll_push: Callable
) -> Result:
    continued_flag = False
    request_headers = request.headers.copy()
    request.save_path.parent.mkdir(exist_ok=True, parents=True)
    if request.save_path.exists():
        request_headers['Range'] = f'bytes={request.save_path.stat().st_size}-'
        request.stat.push(request.save_path.stat().st_size)

    with client.stream(
            method=request.method,
            url=request.url,
            params=request.params,
            data=request.data,
            headers=request_headers,
            cookies=request.cookies,
            auth=request.auth,
            timeout=request.timeout,
            follow_redirects=request.follow_redirects,
    ) as response:
        response: Response
        if response.status_code == 416:
            check_range_transborder(
                client,
                request
            )
            logger.info('The file already exists')
            return Result(
                status=Status.EXIST,
                request=request
            )
        response.raise_for_status()
        if response.headers.get('Accept-Ranges', None) == 'bytes' or response.status_code == 206:
            continued_flag = True
        correct_size = int(response.headers.get('Content-Length', 0))
        request.save_path.parent.mkdir(exist_ok=True, parents=True)
        if correct_size:
            request.correct_size = correct_size
            if continued_flag and correct_size > request.slice_threshold and not request.stream:
                return Result(
                    status=Status.SLICE,
                    request=request
                )
        with request.save_path.open('ab' if continued_flag else 'wb') as f:
            for chunk in response.iter_bytes(chunk_size=request.stream_size):
                f.write(chunk)
                request.stat.push(len(chunk))

        logger.info(f'{request.save_path.name} file download success')
        result = Result(
            status=Status.SUCCESS,
            request=request
        )
        request.success_callback(result)
        return result


def slice_downloader(
        client: Client,
        request: 'Request',
        poll_push: Callable
) -> Result:
    if not request.correct_size:
        raise NotSliceSizeError(request.save_path)
    if request.correct_size / request.slice_size > 100:
        warnings.warn(
            f'{request.save_path.name} if the number of slices in the '
            f'file is too large, check whether the slices are too small')
    slice_flag_name = request.save_path.name.replace('.', '-')
    future_list = []
    for start in range(0, request.correct_size, request.slice_size):
        end = start + request.slice_size - 1
        chunk_path = request.save_path.parent.joinpath(
            '{}--{}.ydstf'.format(
                slice_flag_name,
                start
            )
        )
        future = poll_push(
            chunk_downloader,
            client,
            request,
            start,
            end,
            chunk_path
        )
        future_list.append(future)
    result = [future.result() for future in future_list]
    status = [res[0] for res in result]
    if all(status):
        merge_file(slice_flag_name, request.save_path)
        result = Result(
            status=Status.SUCCESS,
            request=request
        )
        request.success_callback(result)
        return result
    raise list(filter(lambda x: x[1], result))[0]


def chunk_downloader(
        client: Client,
        request: 'Request',
        start: int,
        end: int,
        save_path: Path,
) -> (bool, None | Exception):
    headers = request.headers.copy()
    headers['Range'] = f'bytes={start}-{end}'
    if save_path.exists():
        if save_path.stat().st_size == request.slice_size:
            logger.info('The file chunk already exists skip')
            return True, None
        elif save_path.stat().st_size > request.slice_size:
            save_path.unlink()
        else:
            chunk_start_size = start + save_path.stat().st_size
            headers['Range'] = f'bytes={chunk_start_size}-{end}'
            request.stat.push(save_path.stat().st_size)
            if chunk_start_size == request.correct_size:
                return True, None
    try:
        with client.stream(
                method=request.method,
                url=request.url,
                params=request.params,
                data=request.data,
                headers=headers,
                cookies=request.cookies,
                auth=request.auth,
                timeout=request.timeout,
                follow_redirects=request.follow_redirects,
        ) as response:
            response: Response
            response.raise_for_status()
            with save_path.open('ab') as f:
                for chunk in response.iter_bytes(chunk_size=request.stream_size):
                    f.write(chunk)
                    request.stat.push(len(chunk))
            logger.info(f'{save_path.name} chunk download success')
            return True, None
    except Exception as e:
        logger.error(f'Chunk download error {e}')
        return False, e


async def async_stream_downloader(
        client: AsyncClient,
        request: 'Request',
        semaphore: asyncio.Semaphore
) -> Result:
    continued_flag = False
    request_headers = request.headers.copy()
    request.save_path.parent.mkdir(exist_ok=True, parents=True)
    if request.save_path.exists():
        request_headers['Range'] = f'bytes={request.save_path.stat().st_size}-'
        await request.stat.apush(request.save_path.stat().st_size)
    async with semaphore:
        async with client.stream(
                method=request.method,
                url=request.url,
                params=request.params,
                data=request.data,
                headers=request_headers,
                cookies=request.cookies,
                auth=request.auth,
                timeout=request.timeout,
                follow_redirects=request.follow_redirects,
        ) as response:
            response: Response
            if response.status_code == 416:
                await async_check_range_transborder(
                    client,
                    request
                )
                logger.info('The file already exists')
                return Result(
                    status=Status.EXIST,
                    request=request
                )
            response.raise_for_status()
            if response.headers.get('Accept-Ranges', None) == 'bytes' or response.status_code == 206:
                continued_flag = True
            correct_size = int(response.headers.get('Content-Length', 0))
            request.save_path.parent.mkdir(exist_ok=True, parents=True)
            if correct_size:
                request.correct_size = correct_size
                if continued_flag and correct_size > request.slice_threshold and not request.stream:
                    return Result(
                        status=Status.SLICE,
                        request=request
                    )
            async with aiofiles.open(request.save_path, 'ab' if continued_flag else 'wb') as f:
                async for chunk in response.aiter_bytes(chunk_size=request.stream_size):
                    await f.write(chunk)
                    await request.stat.apush(len(chunk))

            logger.info(f'{request.save_path.name} file download success')
            result = Result(
                status=Status.SUCCESS,
                request=request
            )
            await request.asuccess_callback(result)
            return result


async def async_slice_downloader(
        client: AsyncClient,
        request: 'Request',
        semaphore: asyncio.Semaphore
) -> Result:
    if not request.correct_size:
        raise NotSliceSizeError(request.save_path)
    if request.correct_size / request.slice_size > 100:
        warnings.warn(
            f'{request.save_path.name} if the number of slices in the '
            f'file is too large, check whether the slices are too small')
    slice_flag_name = request.save_path.name.replace('.', '-')
    task_list = []
    for start in range(0, request.correct_size, request.slice_size):
        end = start + request.slice_size - 1
        chunk_path = request.save_path.parent.joinpath(
            '{}--{}.ydstf'.format(
                slice_flag_name,
                start
            )
        )
        task_list.append(
            asyncio.create_task(
                async_chunk_downloader(
                    client,
                    request,
                    start,
                    end,
                    chunk_path,
                    semaphore
                )
            )
        )
    result = await asyncio.gather(*task_list)
    status = [res[0] for res in result]
    if all(status):
        await async_merge_file(slice_flag_name, request.save_path)
        result = Result(
            status=Status.SUCCESS,
            request=request
        )
        await request.asuccess_callback(result)
        return result
    raise list(filter(lambda x: x[1], result))[0]


async def async_chunk_downloader(
        client: AsyncClient,
        request: 'Request',
        start: int,
        end: int,
        save_path: Path,
        semaphore: asyncio.Semaphore
) -> (bool, None | Exception):
    headers = request.headers.copy()
    headers['Range'] = f'bytes={start}-{end}'
    if save_path.exists():
        if save_path.stat().st_size == request.slice_size:
            logger.info('The file chunk already exists skip')
            return True, None
        elif save_path.stat().st_size > request.slice_size:
            save_path.unlink()
        else:
            chunk_start_size = start + save_path.stat().st_size
            headers['Range'] = f'bytes={chunk_start_size}-{end}'
            if chunk_start_size == request.correct_size:
                return True, None
    try:
        async with semaphore:
            async with client.stream(
                    method=request.method,
                    url=request.url,
                    params=request.params,
                    data=request.data,
                    headers=headers,
                    cookies=request.cookies,
                    auth=request.auth,
                    timeout=request.timeout,
                    follow_redirects=request.follow_redirects,
            ) as response:
                response: Response
                response.raise_for_status()
                async with aiofiles.open(save_path, 'ab') as f:
                    async for chunk in response.aiter_bytes(chunk_size=request.stream_size):
                        await f.write(chunk)
                        await request.stat.apush(len(chunk))
                logger.info(f'{save_path.name} chunk download success')
                return True, None
    except Exception as e:
        logger.error(f'Chunk download error {e}')
        return False, e



import asyncio
import threading
import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from yundownload.request import Request


class Stat:
    def __init__(self, request: 'Request'):
        self._request = request
        self._start_time = time.time()
        self._end_time = None
        self._download_size = 0
        self._last_size = 0
        self._last_time = time.time()
        self._lock = threading.Lock()
        self._alock = asyncio.Lock()

    def close(self):
        self._end_time = time.time()

    def push(self, download_size):
        with self._lock:
            self._download_size += download_size

    async def apush(self, download_size):
        async with self._alock:
            self._download_size += download_size

    def clear(self):
        self._download_size = 0

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time

    @property
    def percentage(self):
        if self._request.correct_size is None:
            return 0
        return round(self._download_size / self._request.correct_size, 2)

    @property
    def speed(self):
        now_time = int(time.time())
        speed = (self._download_size - self._last_size) / (now_time - self._last_time)
        self._last_size = self._download_size
        self._last_time = now_time
        return speed
